from typing import Set, Callable, List, Any, Dict, Tuple

import networkx as nx
from evrouting.T import Node, SoC
from evrouting.graph_tools import (
    CONSUMPTION_KEY,
    DISTANCE_KEY,
    consumption,
    distance,
    charging_cofficient
)

Path = List[Node]
DistFunction = Callable[[nx.Graph, Node, Node, str], Path]


def dijkstra(G: nx.Graph, u: Any, v: Any, weight: str = 'weight') -> list:
    return nx.algorithms.shortest_path(G, u, v, weight=weight)


def fold_path(G: nx.Graph, path: Path, weight: str):
    return sum([G.edges[u, v][weight] for u, v in zip(path[:-1], path[1:])])


def insert_start_node(s: Node,
                      graph_core: nx.Graph,
                      graph_contracted: nx.Graph,
                      gas_stations: Set[Node],
                      graph_extended: nx.DiGraph,
                      capacity: SoC,
                      initial_soc: SoC,
                      dist: DistFunction = dijkstra
                      ) -> nx.DiGraph:
    """Insert s into extended graph an create states and edges as necessary."""
    for v in gas_stations:
        shortest_p = dist(graph_core, s, v, weight=CONSUMPTION_KEY)
        w = fold_path(graph_core, shortest_p, weight=CONSUMPTION_KEY)
        if w > initial_soc:
            continue

        d = fold_path(graph_core, shortest_p, weight=DISTANCE_KEY)
        c_v = charging_cofficient(graph_core, v)
        g = initial_soc - w

        graph_extended.add_edge((s, initial_soc), (v, g), weight=d)
        for u in graph_contracted.neighbors(v):
            c_u = charging_cofficient(graph_contracted, u)
            w_v_u = consumption(graph_contracted, u, v)
            d_v_u = distance(graph_contracted, u, v)
            if c_v < c_u:
                graph_extended.add_edge(
                    (v, g),
                    (u, capacity - w_v_u),
                    weight=(capacity - g) * c_v + d_v_u
                )
            elif g < w_v_u:
                graph_extended.add_edge(
                    (v, g),
                    (u, 0),
                    weight=(w_v_u - g) * c_v + d_v_u
                )

    return graph_extended


def insert_final_node(t: Node,
                      graph_core: nx.Graph,
                      gas_stations: Set[Node],
                      graph_extended: nx.DiGraph,
                      capacity: SoC,
                      final_soc: SoC,
                      dist: DistFunction = dijkstra
                      ) -> nx.DiGraph:
    """Insert terminal node into extended graph an create states and edges as necessary."""
    for u in gas_stations:
        shortest_p = dist(graph_core, t, u, weight=CONSUMPTION_KEY)
        w = fold_path(graph_core, shortest_p, weight=CONSUMPTION_KEY)
        if w + final_soc > capacity:
            continue

        d_u_t = fold_path(graph_core, shortest_p, weight=DISTANCE_KEY)
        c_u = charging_cofficient(graph_core, u)
        for g in [g for n, g in graph_extended.nodes if n == u]:
            if g > w + final_soc:
                continue
            graph_extended.add_edge(
                (u, g),
                (t, final_soc),
                weight=(w + final_soc - g) * c_u + d_u_t
            )

    return graph_extended


def contract_graph(G: nx.Graph, charging_stations: Set[Node], capacity: SoC,
                   dist: DistFunction = dijkstra) -> nx.Graph:
    """
    :param G: Original graph
    :param charging_stations: Charging stations
    :param capacity: Maximum battery capacity
    :param dist: Minimum distance function, necessary if G is not a fully
        connected Graph to calculate consumptions between charging stations.
    :returns: Graph only consisting of Charging Stations whose neighbours must
        be within the capacity U. If so, their edge has consumption and
        distance of the minimum path.
    """
    H: nx.Graph = nx.Graph()

    for cs in list(charging_stations):
        H.add_node(cs, **G.nodes[cs])
        # Iterate unvisited charging stations
        for n_cs in [n for n in charging_stations if (n, cs) not in H.edges and n != cs]:
            min_path: Path = dist(G, cs, n_cs)
            consumption: SoC = fold_path(G, min_path, weight=CONSUMPTION_KEY)
            if consumption <= capacity:
                H.add_edge(
                    cs, n_cs,
                    **{
                        CONSUMPTION_KEY: consumption,
                        DISTANCE_KEY: fold_path(G, min_path, weight=DISTANCE_KEY)
                    }
                )
    return H


def get_possible_arriving_soc(G: nx.Graph, u: Node, capacity: SoC) -> List[SoC]:
    """
    :returns: All possible SoC  when arriving at node u, according to
        the optimal fuelling strategy.
    """
    possible_arriving_soc: Set[SoC] = {0}
    c_u = charging_cofficient(G, u)

    for n in G.neighbors(u):
        arriving_soc = capacity - consumption(G, u, n)
        if arriving_soc > 0 and charging_cofficient(G, n) < c_u and \
                arriving_soc not in possible_arriving_soc:
            possible_arriving_soc.add(arriving_soc)

    return list(possible_arriving_soc)


def state_graph(G: nx.Graph, capacity: SoC) -> nx.DiGraph:
    """Calculate Graph connecting (Node, Arrival SoC) states."""
    H: nx.DiGraph = nx.DiGraph()

    for u in G.nodes:
        c_u = charging_cofficient(G, u)
        for v in G.neighbors(u):
            w = consumption(G, u, v)
            if w <= capacity:
                for g in get_possible_arriving_soc(G, u, capacity):
                    c_v = charging_cofficient(G, v)
                    if c_v <= c_u and g < w:
                        weight = (w - g) * c_u + distance(G, u, v)
                        H.add_edge((u, g), (v, 0), weight=weight)
                    elif c_v > c_u:
                        weight = (capacity - g) * c_u + distance(G, u, v)
                        H.add_edge((u, g), (v, capacity - w), weight=weight)

    return H


def compose_result(graph_core: nx.Graph, extended_graph: nx.DiGraph, path: List[Tuple[Node, Node]],
                   dist=dijkstra) -> dict:
    charge_path = [(path[0][0], 0)]  # Start Node
    trip_time = sum([extended_graph.edges[u, v]['weight'] for u, v in zip(path[:-1], path[1:])])

    for i in range(1, len(path) - 1):
        u, g_u = path[i]
        v, g_v = path[i + 1]
        t = extended_graph.edges[(u, g_u), (v, g_v)]['weight']
        charge_time_u = t - fold_path(
            graph_core,
            dist(graph_core, u, v, weight=DISTANCE_KEY),
            weight=DISTANCE_KEY
        )
        charge_path.append((u, charge_time_u))

    charge_path.append((path[-1][0], 0))  # Final Node

    return {'trip_time': trip_time, 'path': charge_path}


def shortest_path(G: nx.Graph, charging_stations: Set[Node], s: Node, t: Node,
                  initial_soc: SoC, final_soc: SoC, capacity: SoC) -> Dict:
    """
    Calculates shortest path using a generalized gas station algorithm.

    :param G: Input Graph
    :param s: Start Node identifier
    :param t: End Node identifier
    :param b_0: Start SoC
    :param b_t: End SoC
    :param U: Capacity
    :return:
    """
    # Check if t is reachable from s
    # Todo: Test!
    _path = dijkstra(G, s, t, weight=CONSUMPTION_KEY)
    _w = fold_path(G, _path, weight=CONSUMPTION_KEY)
    if _w <= initial_soc:
        return {
            'trip_time': fold_path(G, _path, weight=DISTANCE_KEY),
            'path': [(s, 0), (t, 0)]
        }

    contracted_graph: nx.Graph = contract_graph(G, charging_stations, capacity)
    extended_graph = state_graph(contracted_graph, capacity)

    extended_graph = insert_start_node(
        s=s,
        graph_core=G,
        graph_contracted=contracted_graph,
        gas_stations=charging_stations,
        graph_extended=extended_graph,
        capacity=capacity,
        initial_soc=initial_soc
    )

    extended_graph = insert_final_node(
        t=t,
        graph_core=G,
        gas_stations=charging_stations,
        graph_extended=extended_graph,
        capacity=capacity,
        final_soc=final_soc
    )

    path: List[Tuple[Node, Node]] = dijkstra(extended_graph, (s, initial_soc), (t, final_soc))

    return compose_result(
        graph_core=G,
        extended_graph=extended_graph,
        path=path
    )
